/*
 * OpenEAS Experlog Application Server
 * Web application development and dynamic template scripting engine
 * Copyright (c) 2004 ExperLog SARL - http://www.experlog.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version. This program is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have
 * received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */


package org.experlog.db;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Connection;
import java.sql.Statement;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.experlog.base.OpenClient;


/**
 * @author Pierre-Yves Gibello
 *
 */

class Cdata {
  Object value_ = null;
  long wversion_ = 0L;
  
  public Cdata(Object value, long wversion) {
    value_ = value;
    wversion_ = wversion;
  }
  
  public Object value() { return value_; }
  public long version() { return wversion_; }
  
  public String toString() { return (value_ == null ? "null" : value_.toString()); }
}

/**
 * create table ESHashtable (
 *  tabid varchar(32),
 *  ckey varchar(64),
 *  cvalue blob,
 *  wversion integer default 0,
 *  wtime double default 0,
 *  primary key(tabid, ckey));
 */  
public class ESPersistentHashtable {

  private String dbtable_ = "ESHashtable";
  private String tabId_ = "default";
  private OpenClient session_ = null;
  private boolean cache_ = true;
  private String datasource_ = null;
  private boolean singleApp_ = false; // multiple webapps or appserver instances supported
  
  private Hashtable tuple_ = null;
  
  public ESPersistentHashtable(OpenClient session, String dsn, String tabid, String tabname) {
    session_ = session;
    datasource_ = dsn;
    if(tabname != null) dbtable_ = tabname;
    if(tabid != null) {
      tabId_ = tabid;
    } else if(session != null) {
      tabId_ = session.getClientId();
      try {
        cache_ = ! "false".equalsIgnoreCase(session.getAppConfiguration().get("app.persist.cache"));
        singleApp_ = "true".equalsIgnoreCase(session.getAppConfiguration().get("app.persist.single"));
      } catch(Exception ignore) {
        cache_ = true;
        singleApp_ = false;
      }
    }

    // Populate tuple if possible
    if(cache_) populateTuple(false);
  }
  
  private void populateTuple(boolean force) {
    
    if(session_ != null) {

      if(!force && singleApp_ && tuple_ != null) return; // Populate already done
      if(force) tuple_ = null;
      
      Connection c = null;
      try {
        c = session_.getConnection(datasource_);
        Statement st = c.createStatement();
        ResultSet rs = st.executeQuery(
            "select ckey,cvalue,wversion from " + dbtable_ + " where tabid=\'" + tabId_ + "\'");
        while(rs.next()) {
          if(tuple_ == null) tuple_ = new Hashtable(16);
          ObjectInputStream in = new ObjectInputStream(
              new ByteArrayInputStream(rs.getBytes(2)));
          if(tuple_ == null) tuple_ = new Hashtable(16);
          tuple_.put(rs.getString(1), new Cdata(in.readObject(), rs.getLong(3)));
          in.close();
        }
        rs.close();
        st.close();
      } catch(Exception e) {
        if(tuple_ != null && tuple_.size() <= 0) tuple_ = null;
        e.printStackTrace(System.err);
      } finally {
        if(c != null)
          try { session_.closeConnection(c, datasource_);
          } catch(Exception ignore) { }
      }
    }
  }
  
  public void put(String key, Object val) throws SQLException {
    if(tuple_ == null) tuple_ = new Hashtable();
    long wtime = System.currentTimeMillis();
    if(session_ != null) {
      
      synchronized (session_) { // prevent parallel invocation on the same server
        Connection c = null;
        try {
          c = session_.getConnection(datasource_);
          ByteArrayOutputStream bos = new ByteArrayOutputStream();
          ObjectOutputStream out = new ObjectOutputStream(bos);
          out.writeObject(val);
          out.close();
          byte[] ser = bos.toByteArray();

          PreparedStatement st = null;
          //String updateRequest = "update " + dbtable_ + " set cvalue=?,wtime=" + wtime
           // + " where tabid=\'" + tabId_
           // + "\' and ckey=\'" + session_.sqlEncode(datasource_, key) + "\'";
          
          if(cache_ && tuple_.get(key) != null) {
            
            // If cache is disabled, let's consider the "update" is the least frequent case...
            st = c.prepareStatement("update " + dbtable_ + " set cvalue=?,wversion=wversion+1,wtime=" + wtime
                + " where tabid=\'" + tabId_
                + "\' and ckey=\'" + session_.sqlEncode(datasource_, key) + "\'");
                
          } else {
            st = c.prepareStatement("insert into " + dbtable_
                + "(tabid,ckey,cvalue,wversion,wtime) VALUES(\'"
                + tabId_ + "\',"
                + "\'" + session_.sqlEncode(datasource_, key) + "\',"
                + "?,"
                + "1," // version 1 at 1st insert
                + wtime
                + ")");
          }
          
          try {
            st.setBytes(1, ser);
            st.executeUpdate();
            st.close();
          } catch(SQLException e) {
            // A DUPLICATE KEY error may occur in case of parallel invocation from 2 servers.
            // There is no way to prevent it (eg. "synchronized") because there are 2 JVMs involved...
            // So, if an exception occurs, assume the PK was already there... and update the value.
            if(st != null) {
              try { st.close(); } catch(Exception ignore) { }
            }
            //st = c.prepareStatement(updateRequest);
            
            // Here we arbitrarily add 5 (not just 1) to the version number
            // This ensures more than 2 concurrent calls from different servers just increment it by 1
            // It is unlikely that 6 parallel calls from 6 different servers can occur...
            st = c.prepareStatement("update " + dbtable_ + " set cvalue=?,wversion=wversion+5,wtime=" + wtime
                + " where tabid=\'" + tabId_
                + "\' and ckey=\'" + session_.sqlEncode(datasource_, key) + "\'");
            st.setBytes(1, ser);
            st.executeUpdate();
            st.close();
          }
          
        } catch(SQLException e) {
          //e.printStackTrace(System.err);
          throw e;
        } catch(Exception e) {
          throw new SQLException(e.getClass() + ": " + e.getMessage());
        } finally {
          if(c != null)
            try { session_.closeConnection(c, datasource_);
            } catch(Exception ignore) { }
        }
      } // end synchronized block
      
    } else { // No persistence (session is null)

      //tuple_.put(key, new Cdata(val, wtime)); (TBD... when wtime was a timestamp)
      tuple_.put(key, new Cdata(val, 0));
    }
  }

  public void putNoPersist(String key, Object val) {
    if(tuple_ == null) tuple_ = new Hashtable();
    tuple_.put(key, new Cdata(val, -1)); // version=-1 for non-persistent data
  }
  
  public Object get(String key) throws SQLException {
    long version = 0;
    
    if(cache_ && tuple_ == null) {
      populateTuple(false);
      if(tuple_ == null) return null;
    }

    Cdata data = (tuple_ == null ? null : (Cdata)tuple_.get(key));
    Object ret = (data == null ? null : data.value());
    
    if(data != null && data.version() == -1) return data.value(); // -1 for non-persistent data

    if(session_ != null) {
      version = (data == null ? 0 : data.version());

      Connection c = null;
      try {
        c = session_.getConnection(datasource_);
        Statement st = c.createStatement();
        ResultSet rs = st.executeQuery(
            "select cvalue,wversion from " + dbtable_ + " where tabid=\'" + tabId_
            + "\' and ckey=\'" + session_.sqlEncode(datasource_, key) + "\'"
            + (cache_ ? (" and wversion>" + version) : ""));
        if(rs.next()) { // value changed (or no cache)
          ObjectInputStream in = new ObjectInputStream(
              new ByteArrayInputStream(rs.getBytes(1)));
          ret = in.readObject();
          in.close();
          if(cache_) tuple_.put(key, new Cdata(ret, rs.getLong(2)));
        }
        rs.close();
        st.close();
      } catch(Exception ignore) {
        ignore.printStackTrace(System.err);
      } finally {
        if(c != null)
          try { session_.closeConnection(c, datasource_);
          } catch(Exception ee) { }
      }
    }

    return ret;
  }

  public Enumeration keys() {
    if(cache_ || session_ == null) {
      if(tuple_ == null) return null;
      else return tuple_.keys();
      
    } else {
      Connection c = null;
      Vector v = null;
      try {
        c = session_.getConnection(datasource_);
        Statement st = c.createStatement();
        ResultSet rs = st.executeQuery("select ckey from " + dbtable_ + " where tabid=\'" + tabId_);
        while(rs.next()) {
          if(v == null) v = new Vector();
          v.addElement(rs.getString(1));
        }
        rs.close();
        st.close();
        
      } catch(Exception ignore) {
        ignore.printStackTrace(System.err);
      } finally {
        if(c != null)
          try { session_.closeConnection(c, datasource_);
          } catch(Exception ee) { }
      }
      return (v == null ? null : v.elements());
    }
  }
  
  public void remove(String key) throws SQLException {
    if(cache_ && tuple_ == null) return;
    if(session_ != null) {
      Connection c = null;
      try {
        c = session_.getConnection(datasource_);
        Statement st = c.createStatement();
        st.executeUpdate("delete from " + dbtable_ + " where tabid=\'" + tabId_
            + "\' and ckey=\'" + session_.sqlEncode(datasource_, key) + "\'");
        st.close();
      } catch(Exception ignore) {
        ignore.printStackTrace(System.err);
      } finally {
        if(c != null)
          try { session_.closeConnection(c, datasource_);
          } catch(Exception ee) { }
      }
    }
    if(tuple_ != null) tuple_.remove(key);
  }
  
  public void clear() {
    /** TBD if persistent do not clear DB !!!
    if(session_ != null) {
      Connection c = null;
      try {
        c = session_.getConnection(datasource_);
        Statement st = c.createStatement();
        st.executeUpdate("delete from " + dbtable_ + " where tabid=\'" + tabId_ + "\'");
        st.close();
      } catch(Exception ignore) {
        ignore.printStackTrace(System.err);
      } finally {
        if(c != null)
          try { session_.closeConnection(c, datasource_);
          } catch(Exception ee) { }
      }
    } **/
    
    if(tuple_ != null) tuple_.clear();
  }

  public void cleanup(int days) {
    if(session_ != null) {
      Connection c = null;
      try {
        c = session_.getConnection(datasource_);
        Statement st = c.createStatement();
        long old = System.currentTimeMillis() - (long)days*24*3600*1000;
        st.executeUpdate("delete from " + dbtable_ + " where wtime<" + old
            + " and tabid=\'" + tabId_ + "\'");
        st.close();
      } catch(Exception ignore) {
        ignore.printStackTrace(System.err);
      } finally {
        if(c != null)
          try { session_.closeConnection(c, datasource_);
          } catch(Exception ee) { }
      }
    }
  }
  
  public Hashtable toHashtable() {
    if(! cache_) populateTuple(true);
    return tuple_;
  }

  public void fromHashtable(Hashtable t) throws SQLException {
    clear();
    Enumeration keys = t.keys();
    while(keys.hasMoreElements()) {
      String k = (String)keys.nextElement();
      put(k, t.get(k));
    }
  }

  /* TBD
  public String[] insertQuery(Hashtable tuple) {
    
    String ret[] = new String[tuple.size()];
    int k = 0;
    
    Enumeration en = tuple.keys();
    while (en.hasMoreElements()) {
      String key = (String) en.nextElement();
      Object value = tuple.get(key);
      if(value != null) {

        ret[k++] = new String("INSERT INTO " + dbtable_ + "(tabid,ckey,cvalue,wtime) VALUES(\'" + tabId_ + "\',"
          + "\'" + session_.sqlEncode(datasource_, key) + "\',?,"
          + System.currentTimeMillis()
          + ")");
      }
    }

    return ret;
  }
  */

  /* TBD
  public static void main(String args[]) {
  try {
      ESPersistentHashtable dbo =
        new ESPersistentHashtable(null, "test");
      dbo.addColumn("Id", Types.VARCHAR, "001");
      dbo.setWritable("Id", false);
      dbo.addColumn("FirstName", Types.VARCHAR, "Bill");
      dbo.addColumn("LastName", Types.VARCHAR, "Smith");
      dbo.addColumn("Age", Types.INTEGER, "30");

      dbo.set("FirstName", "John");
      System.out.println(dbo.updateQuery("Id=\'001\'"));
      System.out.println(dbo.insertQuery());
      System.out.println(dbo.deleteQuery(null));
  } catch (Exception e) {
      System.err.println("ESDbObject:main Exception " + e);
  }
  }
  */

}
