/*
 * OpenEAS Experlog Application Server
 * Web application development and dynamic template scripting engine
 * Copyright (c) 2004 ExperLog SARL - http://www.experlog.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version. This program is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have
 * received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.experlog.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.StringReader;
import java.net.InetAddress;
import java.net.Socket;
import java.util.StringTokenizer;

public class ESMailer {

	String mailhost, helohost;
	String result = "";
	String subject = null;
	String message;
	String lastline;
	int port = 25;
	BufferedReader in;

	String ctype_ = null;

	boolean legacy_ = false;

	public ESMailer(String mhost) {

		try {
			if (mhost == null)
				mailhost = InetAddress.getLocalHost().getHostAddress();
			else
				mailhost = mhost;
		} catch (Exception e) {
			System.err.println("ESMailer Exception " + e);
		}
	}

	void expect(String expected, String msg) throws IOException {

		lastline = in.readLine();

		if (!lastline.startsWith(expected))
			throw new IOException(msg + ":" + lastline);

		while (lastline.startsWith(expected + "-"))
			lastline = in.readLine();

	}

	public void setPort(int p) {
		port = p;  // If not 25, for any reason...
	}

	public void setContentType(String ctype) {
		ctype_ = ctype;
	}

	public Socket open() throws IOException {
		return new Socket(mailhost, port);
	}

	public PrintWriter beginMessage(Socket s, String from, String to,
									String subject) throws IOException {

		if (subject == null) subject = "(no subject)";

		PrintWriter p = new PrintWriter(new BufferedWriter(
				new OutputStreamWriter(s.getOutputStream())), true);

		in = new BufferedReader(new InputStreamReader(
				s.getInputStream()));

		expect("220", "no greeting");

		helohost = InetAddress.getLocalHost().getHostAddress();
		p.println("HELO " + helohost + "\r");
		expect("250", "helo");

		p.println("MAIL FROM: <" + from + ">\r");
		expect("250", "mail from");

		// Handle multiple recipients, comma-separated
		StringTokenizer st = new StringTokenizer(to, ",");
		while (st.hasMoreTokens()) {
			p.println("RCPT TO: <" + st.nextToken().trim() + ">\r");
			expect("250", "rcpt to");
		}

		p.println("DATA" + "\r");
		expect("354", "data");

		if(legacy_) {
			p.println("Subject: " + subject + "\r");
		} else {
			p.println("Subject: " + subject);
			p.println();
		}

		if (ctype_ != null) {
			p.print("Content-type: " + ctype_);
			p.println(ctype_.endsWith(";") ? "" : ";");
			p.println("\r");
		}

		return p;
	}

	public void endMessage(PrintWriter p) throws IOException {
		if(legacy_) {
			p.println("\r\n.\r");
		} else {
			// Print a <CR><LF>.<CR><LF>
			// Trusting println() for that seems to be the best option,
			// on both Win32 and Unix...
			p.println();
			p.print(".");
			p.println(); p.flush();
		}
		expect("250", "end of data");

		p.println("QUIT\r");
		expect("221", "quit");
	}

	public void close(Socket s) throws IOException {
		s.close();
	}

	public void setLegacy(boolean l) { legacy_ = l; }

	public boolean sendMail(String from, String to, String subject,
							BufferedReader msgreader) {

		Socket s = null;
		try {

			if (subject == null) subject = "(no subject)";

			s = open();

			PrintWriter p = beginMessage(s, from, to, subject);

			String ln;
			while ((ln = msgreader.readLine()) != null) {
				if (ln.equals("."))
					ln = "..";
				p.println(ln + "\r");
			}

			endMessage(p);
			close(s);

		} catch (Exception e) {
			System.err.println("ESMailer:sendMail Exception " + e);
			result = e.getMessage();
			return false;
		} finally {
			try {
				if (s != null)
					close(s);
			} catch (Exception e) {
				System.err.println("ESMailer:sendMail Exception " + e);
				result = e.getMessage();
			}
		}

		return true;
	}

	public boolean sendMail(String from, String to, String subject, Object o) {
		return sendMail(from, to, subject,
						new BufferedReader(new StringReader(o.toString())));
	}

	/**
	 * Test program:
	 * java ESMailer to [mailhost] [from]
	 */
	public static void main(String args[]) {

		try {
			if(args.length <= 0) {
				System.err.println("Usage: java ESMailer to [mailhost] [from]");
				System.exit(1);
			}
			String host = (args.length > 1 ? args[1] : null);
			String from = (args.length > 2 ? args[2] : "test@experlog.com");
			if(args.length == 2 && host.indexOf('@') > 0) {
				from = host;
				host = null;
			}

			boolean legacy = false;
			if(host != null && host.startsWith("[legacy]")) {
				host = host.substring(8);
				legacy = true;
			}

			System.out.println("host: " + host);
			System.out.println("from: " + from);
			System.out.println("to: " + args[0]);
			if(legacy) System.out.println("legacy flag is set...");

			ESMailer m = new ESMailer(host);
			m.setLegacy(legacy);

			if (m.sendMail(from, args[0], "ESMailer test program",
						   new BufferedReader(new StringReader("Seems to work..."))) == false)
				System.err.println("Error while sending mail (no exception thrown).");

		} catch (Exception e) {
			System.err.println("ESMailer Exception: " + e);
			e.printStackTrace(System.err);
		}
	}
}

